"""Module for storing client requests when server not available, and recovering
of such.
"""

import json
import os.path

from financeager import init_logger

from . import OFFLINE_FILEPATH, exceptions

logger = init_logger(__name__)


def _load(filepath):
    if os.path.exists(filepath):
        with open(filepath, "r") as file:
            content = json.load(file)
            logger.debug("Loaded {}".format(content))
    else:
        content = []

    return content


def _write(content, filepath):
    with open(filepath, "w") as file:
        logger.debug("Writing {}".format(content))
        json.dump(content, file)


def _recover_data(client, content):
    """Recover the data items in the content list by running the commands via
    'client.safely_run'. If an error occurs, the recovery is aborted, and the
    currently processed data is returned.
    The content list is modified in-place.
    """
    while len(content):
        data = content.pop()
        if not client.safely_run(**data):
            return data


def add(command, offline_filepath=None, **cl_kwargs):
    """Add a command and optional kwargs passed from the command line to the
    offline backup database.

    Non-modifying request commands such as 'print' or 'list' are not stored.

    :return: if anything was added
    """

    if command not in ["add", "remove", "update"]:
        return False

    offline_filepath = offline_filepath or OFFLINE_FILEPATH
    content = _load(offline_filepath)

    cl_kwargs["command"] = command
    content.append(cl_kwargs)
    _write(content, offline_filepath)

    return True


def recover(client, offline_filepath=None):
    """Recover the offline backup by passing its content to the given client.
    The recovery will be aborted if a CommunicationError occurs.

    If the recovery succeeded, the backup file is deleted.

    :return: if anything was recovered
    :raises: OfflineRecoveryError if recovery failed
    """

    offline_filepath = offline_filepath or OFFLINE_FILEPATH

    content = _load(offline_filepath)
    if not content:
        return False

    failed_recovery_data = _recover_data(client, content)

    if failed_recovery_data is None:
        os.remove(offline_filepath)
    else:
        content.append(failed_recovery_data)
        _write(content, offline_filepath)
        raise exceptions.OfflineRecoveryError()

    return True



"""Utiliary classes for testing."""
from financeager import clients

from financeager_flask import main


class Client(clients.LocalServerClient):
    """Implementation that assigns dummy sinks to consume the client's output.
    The underlying Proxy is patched to store data in memory instead of in
    financeager.DATA_DIR.
    """

    def __init__(self):
        f = lambda s: None
        super().__init__(
            configuration=main._Configuration(), sinks=clients.Client.Sinks(f, f)
        )

        self.proxy._pocket_kwargs["data_dir"] = None



import os.path
import unittest

from financeager_flask import exceptions
from financeager_flask.offline import _load, add, recover

from . import utils


class AddTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filepath = os.path.join(os.path.expanduser("~"), "offline_test.json")

    def test_add_recover(self):
        pocket_name = "123"
        kwargs = dict(name="money", value=111, date="2019-01-31", pocket=pocket_name)
        self.assertTrue(add("add", offline_filepath=self.filepath, **kwargs))

        content = _load(self.filepath)

        self.assertIsInstance(content, list)
        self.assertEqual(len(content), 1)
        data = content[0]
        self.assertEqual(data.pop("command"), "add")
        self.assertDictEqual(kwargs, data)

        client = utils.Client()
        self.assertTrue(recover(client, offline_filepath=self.filepath))

        element = client.proxy.run("get", eid=1, pocket=pocket_name)["element"]
        self.assertEqual(element["name"], "money")
        self.assertEqual(element["value"], 111)

    def test_no_add(self):
        self.assertFalse(add("list"))

    def test_no_recover(self):
        self.assertFalse(recover(None, offline_filepath=self.filepath))

    def test_failed_recover(self):
        pocket_name = "123"
        kwargs = dict(name="money", value=111, pocket=pocket_name)
        command = "add"
        self.assertTrue(add(command, offline_filepath=self.filepath, **kwargs))
        self.assertTrue(add(command, offline_filepath=self.filepath, **kwargs))

        # Create client, and fake the run method
        client = utils.Client()

        def run(command, **kwargs):
            raise Exception

        client.proxy.run = run

        self.assertRaises(
            exceptions.OfflineRecoveryError,
            recover,
            client,
            offline_filepath=self.filepath,
        )

        content = _load(self.filepath)
        kwargs["command"] = command
        self.assertDictEqual(content[0], kwargs)
        self.assertDictEqual(content[1], kwargs)

    def tearDown(self):
        if os.path.exists(self.filepath):
            os.remove(self.filepath)


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from os import environ
from unittest import mock

import financeager

from financeager_flask.flask import create_app

# Patch DATA_DIR to avoid having it created/interfering with logs on actual
# machine
TEST_DATA_DIR = tempfile.mkdtemp(prefix="financeager-")


@mock.patch("financeager.DATA_DIR", TEST_DATA_DIR)
class CreateAppNoDataDirTestCase(unittest.TestCase):
    @mock.patch("financeager_flask.flask.logger.warning")
    def test_warning(self, mocked_warning):
        create_app()
        mocked_warning.assert_called_once_with(
            "'data_dir' not given. Application data is stored in "
            "memory and is lost when the flask app terminates. Set "
            "the environment variable FINANCEAGER_FLASK_DATA_DIR "
            "accordingly for persistent data storage."
        )

    def test_debug(self):
        app = create_app(config={"DEBUG": True})
        self.assertTrue(app.debug)
        self.assertEqual(financeager.LOGGER.handlers[0].level, 10)

    def test_data_dir_env_variable(self):
        data_dir = tempfile.mkdtemp(prefix="financeager-")
        environ["FINANCEAGER_FLASK_DATA_DIR"] = data_dir
        with mock.patch("os.makedirs") as mocked_makedirs:
            create_app(data_dir=None)
            # First call is from inside setup_log_file_handler()
            self.assertEqual(mocked_makedirs.call_count, 2)
            mocked_makedirs.assert_called_with(data_dir, exist_ok=True)
        del environ["FINANCEAGER_FLASK_DATA_DIR"]

    def test_bad_request(self):
        app = create_app()
        app.testing = True
        with app.test_client() as client:
            response = client.post("/pockets/2000")
        # Expect Bad Request due to missing data (name and value)
        self.assertEqual(response.status_code, 400)


if __name__ == "__main__":
    unittest.main()



#!/home/USER/.virtualenvs/financeager/bin/python3
"""This is an examplatory FCGI script. I use it to run financeager on uberspace.
It is assumed that financeager and the flipflop-package (interface between flask
and FCGI) are installed:
    python3 -m venv ~/.virtualenvs/financeager
    source ~/.virtualenvs/financeager/bin/activate
    pip install -U pip
    pip install financeager-flask flipflop==1.0

The script is placed in ~/fcgi-bin and made executable. Adjust the shebang-line.

On the client side, specify the host as
    http://USER.STAR.uberspace.de/fcgi-bin/financeager.fcgi

The uberspace domain is publicly accessible. For basic authentication, run
    htpasswd -m -c /var/www/virtual/USER/.htuser USER

and add a file ~/fcgi-bin/.htaccess with content
    AuthType Basic
    AuthName flask-financeager
    AuthUserFile /var/www/virtual/USER/.htuser
    Require valid-user

On the client side, specify user and password in the config file as documented
in the main README.

You can prettify the URL format by adding a file ~/html/.htaccess with content
    RewriteEngine On
    RewriteRule ^financeager/(.*)$ /fcgi-bin/financeager.fcgi/$1 [QSA,L]

Force a restart of the FCGI app by killing the existing one.
This allows you to access via
    http://USER.STAR.uberspace.de/financeager

Further reading:
https://blog.lucas-hild.de/flask-uberspace
https://gist.github.com/tocsinDE/98c423da2724d23c02ff
https://docs.python.org/3.4/howto/webservers.html
https://wiki.uberspace.de/webserver:htaccess#verzeichnisschutz
"""
from flipflop import WSGIServer

from financeager_flask import DATA_DIR, flask

if __name__ == "__main__":
    # Configure to your liking
    app = flask.create_app(
        data_dir=DATA_DIR,
        # config={"DEBUG": True},
    )
    WSGIServer(app).run()



"""Webservice resources as end points of financeager REST API."""
import json

import flask
from financeager import exceptions, init_logger
from flask_restful import Resource, reqparse

from . import version

logger = init_logger(__name__)

copy_parser = reqparse.RequestParser()
copy_parser.add_argument("destination_pocket", required=True)
copy_parser.add_argument("source_pocket", required=True)
copy_parser.add_argument("eid", required=True, type=int)
copy_parser.add_argument("table_name")

put_parser = reqparse.RequestParser()
put_parser.add_argument("name", required=True)
put_parser.add_argument("value", required=True, type=float)
put_parser.add_argument("category")
put_parser.add_argument("date")
put_parser.add_argument("frequency")
put_parser.add_argument("start")
put_parser.add_argument("end")
put_parser.add_argument("table_name")

print_parser = reqparse.RequestParser()
print_parser.add_argument("filters")
print_parser.add_argument("recurrent_only", type=bool)

update_parser = reqparse.RequestParser()
update_parser.add_argument("name")
update_parser.add_argument("value", type=float)
update_parser.add_argument("category")
update_parser.add_argument("date")
update_parser.add_argument("frequency")
update_parser.add_argument("start")
update_parser.add_argument("end")


class LogResource(Resource):
    """Custom class to facilitate request logging and safe execution of server
    commands."""

    def __init__(self, server, *args, **kwargs):
        """Store reference to application Server object."""
        super().__init__(*args, **kwargs)
        self.server = server

    def run_safely(self, command, **kwargs):
        """Wrapper function for running commands on server. Returns server
        response, if erroneous, including an appropriate error code.
        If an unexpected exception is caught, the method logs it and returns an
        internal server error.
        """
        try:
            response = self.server.run(command, **kwargs)

            if "error" in response:
                error = response["error"]
                if isinstance(error, exceptions.PocketEntryNotFound):
                    error_code = 404
                else:
                    error_code = 400
                response = ({"error": str(error)}, error_code)

        except Exception:
            logger.exception("Unexpected error")
            response = ({"error": "unexpected error"}, 500)

        return response

    def dispatch_request(self, *args, **kwargs):
        """Log content of request that is about to be dispatched."""
        logger.debug(
            "Dispatching {r} holding {{data: {r.data}, "
            "values: {r.values}, json: {r.json}}}".format(r=flask.request)
        )
        return super().dispatch_request(*args, **kwargs)


class PocketsResource(LogResource):
    def post(self):
        return self.run_safely("pockets")


class PocketResource(LogResource):
    def get(self, pocket_name):
        args = json.loads(flask.request.json or "{}")
        return self.run_safely("list", pocket=pocket_name, **args)

    def post(self, pocket_name):
        args = put_parser.parse_args()
        return self.run_safely("add", pocket=pocket_name, **args)


class EntryResource(LogResource):
    def get(self, pocket_name, table_name, eid):
        return self.run_safely(
            "get", pocket=pocket_name, table_name=table_name, eid=eid
        )

    def delete(self, pocket_name, table_name, eid):
        return self.run_safely(
            "remove", pocket=pocket_name, table_name=table_name, eid=eid
        )

    def patch(self, pocket_name, table_name, eid):
        args = update_parser.parse_args()
        return self.run_safely(
            "update", pocket=pocket_name, table_name=table_name, eid=eid, **args
        )


class CopyResource(LogResource):
    def post(self):
        args = copy_parser.parse_args()
        return self.run_safely("copy", **args)


class VersionResource(LogResource):
    def get(self):
        return {
            "version": version(),
            "financeager_version": version("financeager"),
        }



import os
import tempfile
import time
import unittest
from collections import defaultdict
from threading import Thread
from unittest import mock

from financeager import (
    DEFAULT_TABLE,
    RECURRENT_TABLE,
    cli,
    clients,
    config,
    setup_log_file_handler,
)
from requests import RequestException, Response
from requests import get as requests_get

from financeager_flask import flask, main, version

TEST_CONFIG_FILEPATH = "/tmp/financeager-test-config"
TEST_DATA_DIR = tempfile.mkdtemp(prefix="financeager-")
setup_log_file_handler(log_dir=TEST_DATA_DIR)


class CliTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create test config file for client
        with open(TEST_CONFIG_FILEPATH, "w") as file:
            file.write(cls.CONFIG_FILE_CONTENT)

        cls.pocket = 1900

    def setUp(self):
        # Separate test runs by running individual test methods using distinct
        # pockets (especially crucial for CliFlaskTestCase which uses a single
        # Flask instance for all tests)
        self.__class__.pocket += 1

        # Mocks to record output of cli.run() call
        self.info = mock.MagicMock()
        self.error = mock.MagicMock()

    def cli_run(self, command_line, log_method="info", format_args=()):
        """Wrapper around cli.run() function. Adds convenient command line
        options (pocket and config filepath). Executes the actual run() function
        while patching the module logger info and error methods to catch their
        call arguments.

        'command_line' is a string of the form that financeager is called from
        the command line with. 'format_args' are optional objects that are
        formatted into the command line string. Must be passed as tuple if more
        than one.

        If information about an added/update/removed/copied element was to be
        logged, the corresponding ID is matched from the log call arguments to
        the specified 'log_method' and returned. Otherwise the raw log call is
        returned.
        """
        self.info.reset_mock()
        self.error.reset_mock()

        if not isinstance(format_args, tuple):
            format_args = (format_args,)
        args = command_line.format(*format_args).split()
        command = args[0]

        # Exclude option from subcommand parsers that would be confused
        if command not in ["copy", "pockets", "web-version"]:
            args.extend(["--pocket", str(self.pocket)])

        args.extend(["--config-filepath", TEST_CONFIG_FILEPATH])

        sinks = clients.Client.Sinks(self.info, self.error)

        # Procedure similar to cli.main()
        plugins = [main.main()]
        params = cli._parse_command(args, plugins=plugins)
        configuration = config.Configuration(
            params.pop("config_filepath"), plugins=plugins
        )
        exit_code = cli.run(
            sinks=sinks, configuration=configuration, plugins=plugins, **params
        )

        # Get first of the args of the call of specified log method
        response = getattr(self, log_method).call_args[0][0]

        # Verify exit code
        self.assertEqual(
            exit_code, cli.SUCCESS if log_method == "info" else cli.FAILURE
        )

        # Immediately return str messages
        if isinstance(response, str):
            return response

        # Convert Exceptions to string
        if isinstance(response, Exception):
            return str(response)

        if command in ["add", "update", "remove", "copy"]:
            return response["id"]

        if command in ["get", "pockets"] and log_method == "info":
            return cli._format_response(
                response,
                command,
                default_category=configuration.get_option(
                    "FRONTEND", "default_category"
                ),
                recurrent_only=params.get("recurrent_only", False),
            )

        if command == "list" and log_method == "info":
            return response

        return response


@mock.patch("financeager.DATA_DIR", TEST_DATA_DIR)
class CliFlaskTestCase(CliTestCase):

    HOST_IP = "127.0.0.1:5000"
    CONFIG_FILE_CONTENT = """\
[SERVICE]
name = flask

[FRONTEND]
default_category = unspecified
date_format = %%m-%%d

[SERVICE:FLASK]
host = http://{}
""".format(
        HOST_IP
    )

    @staticmethod
    def launch_server():
        # Patch DATA_DIR inside the thread to avoid having it
        # created/interfering with logs on actual machine
        import financeager

        financeager.DATA_DIR = TEST_DATA_DIR
        app = flask.create_app(
            data_dir=TEST_DATA_DIR,
            config={
                "DEBUG": False,  # reloader can only be run in main thread
                "SERVER_NAME": CliFlaskTestCase.HOST_IP,
            },
        )

        def shutdown():
            from flask import request

            app._server.run("stop")
            request.environ.get("werkzeug.server.shutdown")()
            return ""

        # For testing, add rule to shutdown Flask app
        app.add_url_rule("/stop", "stop", shutdown)

        app.run()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.flask_thread = Thread(target=cls.launch_server)
        cls.flask_thread.start()

        # wait for flask server being launched
        time.sleep(0.2)

    @classmethod
    def tearDownClass(cls):
        # Invoke shutting down of Flask app
        requests_get("http://{}/stop".format(cls.HOST_IP))

    def test_add_list_remove(self):
        entry_id = self.cli_run("add cookies -100")

        response = self.cli_run("list")["elements"]
        self.assertIsNone(response[DEFAULT_TABLE][str(entry_id)]["category"])

        remove_entry_id = self.cli_run("remove {}", format_args=entry_id)
        self.assertEqual(remove_entry_id, entry_id)

        response = self.cli_run("pockets")
        self.assertIn(str(self.pocket), response)

    def test_add_get_remove_via_eid(self):
        entry_id = self.cli_run("add donuts -50 -c sweets")

        response = self.cli_run("get {}", format_args=entry_id)
        name = response.split("\n")[0].split()[2]
        self.assertEqual(name, "Donuts")

        self.cli_run("remove {}", format_args=entry_id)

        response = self.cli_run("list")["elements"]
        self.assertEqual(
            response, {DEFAULT_TABLE: {}, RECURRENT_TABLE: defaultdict(list)}
        )

    def test_add_invalid_entry_table_name(self):
        response = self.cli_run("add stuff 11.11 -t unknown", log_method="error")
        self.assertIn("400", response)

    def test_update(self):
        entry_id = self.cli_run("add donuts -50 -c sweets")

        update_entry_id = self.cli_run("update {} -n bretzels", format_args=entry_id)
        self.assertEqual(entry_id, update_entry_id)

        response = self.cli_run("get {}", format_args=entry_id)
        self.assertIn("Bretzels", response)

    def test_update_nonexisting_entry(self):
        response = self.cli_run("update -1 -n a", log_method="error")
        self.assertIn("404", response)

    def test_get_nonexisting_entry(self):
        response = self.cli_run("get -1", log_method="error")
        self.assertIn("404", response)

    def test_remove_nonexisting_entry(self):
        response = self.cli_run("remove 0", log_method="error")
        self.assertIn("404", response)

    def test_recurrent_entry(self):
        entry_id = self.cli_run(
            "add cookies -10 -c food -t recurrent -f "
            "half-yearly -s 2020-01-01 -e 2020-12-31"
        )
        self.assertEqual(entry_id, 1)

        response = self.cli_run("get {} -t recurrent", format_args=entry_id)
        self.assertIn("Half-Yearly", response)

        update_entry_id = self.cli_run(
            "update {} -t recurrent -n clifbars -f quarter-yearly", format_args=entry_id
        )
        self.assertEqual(update_entry_id, entry_id)

        response = self.cli_run("list")["elements"]
        self.assertEqual(len(response[RECURRENT_TABLE][str(update_entry_id)]), 4)

        self.cli_run("remove {} -t recurrent", format_args=entry_id)

        response = self.cli_run("list")["elements"]
        self.assertEqual(
            response, {DEFAULT_TABLE: {}, RECURRENT_TABLE: defaultdict(list)}
        )

    def test_copy(self):
        destination_pocket = self.pocket + 1
        self.__class__.pocket += 1

        source_entry_id = self.cli_run("add donuts -50 -c sweets")

        destination_entry_id = self.cli_run(
            "copy {} -s {} -d {}",
            format_args=(source_entry_id, self.pocket, destination_pocket),
        )

        # Swap pocket to trick cli_run()
        self.pocket, destination_pocket = destination_pocket, self.pocket
        destination_printed_content = self.cli_run(
            "get {}", format_args=destination_entry_id
        ).splitlines()
        self.pocket, destination_pocket = destination_pocket, self.pocket

        source_printed_content = self.cli_run(
            "get {}", format_args=source_entry_id
        ).splitlines()
        # Remove date lines
        destination_printed_content.remove(destination_printed_content[2])
        source_printed_content.remove(source_printed_content[2])
        self.assertListEqual(destination_printed_content, source_printed_content)

    def test_copy_nonexisting_entry(self):
        destination_pocket = self.pocket + 1
        self.__class__.pocket += 1

        response = self.cli_run(
            "copy 0 -s {} -d {}",
            log_method="error",
            format_args=(self.pocket, destination_pocket),
        )
        self.assertIn("404", response)

    def test_communication_error(self):
        with mock.patch("requests.get") as mocked_get:
            response = Response()
            response.status_code = 500
            mocked_get.return_value = response
            response = self.cli_run("list", log_method="error")
            self.assertIn("500", response)

    @mock.patch(
        "financeager_flask.offline.OFFLINE_FILEPATH",
        os.path.join(TEST_DATA_DIR, "financeager-test-offline.json"),
    )
    def test_offline_feature(self):
        with mock.patch("requests.post") as mocked_post:
            # Try do add an item but provoke CommunicationError
            mocked_post.side_effect = RequestException("did not work")

            self.cli_run("add veggies -33", log_method="error")

            # Output from caught CommunicationError
            self.assertEqual(
                "Error sending request: did not work",
                str(self.error.call_args_list[0][0][0]),
            )
            self.assertEqual(
                "Stored 'add' request in offline backup.",
                self.info.call_args_list[0][0][0],
            )

            # Now request a print, and try to recover the offline backup
            # But adding is still expected to fail
            mocked_post.side_effect = RequestException("still no works")
            self.cli_run("list", log_method="error")
            # Output from print; expect empty database
            self.assertEqual(
                {"elements": {DEFAULT_TABLE: {}, "recurrent": {}}},
                self.info.call_args_list[0][0][0],
            )

            # Output from cli module
            self.assertEqual(
                "Offline backup recovery failed!", self.error.call_args_list[-1][0][0]
            )

        # Without side effects, recover the offline backup
        self.cli_run("list")

        # Output from list command
        self.assertEqual(
            {"elements": {DEFAULT_TABLE: {}, "recurrent": {}}},
            self.info.call_args_list[0][0][0],
        )
        # Output from recovered add command
        self.assertEqual({"id": 1}, self.info.call_args_list[1][0][0])
        self.assertEqual("Recovered offline backup.", self.info.call_args_list[2][0][0])

    def test_web_version(self):
        response = self.cli_run("web-version")
        self.assertIn(version(), response)

    def test_list_recurrent_only(self):
        self.cli_run("add rent -500 -f monthly")
        response = self.cli_run("list --recurrent-only")
        self.assertEqual(response["elements"][0]["name"], "rent")
        self.assertEqual(response["elements"][0]["frequency"], "monthly")


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

from financeager.exceptions import CommunicationError

from financeager_flask import DEFAULT_HOST, POCKETS_TAIL
from financeager_flask.httprequests import Proxy as HttpProxy


class HttpRequestProxyTestCase(unittest.TestCase):
    def mock_post(*args, **kwargs):
        """Mock a Response object returned by requests.post."""

        class MockResponse:
            ok = True

            def json(self):
                return {}

        return MockResponse()

    def test_username_password(self):
        username = "noob"
        password = 123456
        timeout = 1
        proxy = HttpProxy(
            http_config={
                "username": username,
                "password": password,
                "timeout": timeout,
            }
        )

        with patch(
            "financeager_flask.httprequests.requests.post", side_effect=self.mock_post
        ) as post_patch:

            proxy.run("pockets")

            url = "{}{}".format(DEFAULT_HOST, POCKETS_TAIL)
            kwargs = {
                "json": None,
                "auth": (username, password),
                "timeout": timeout,
            }
            post_patch.assert_called_once_with(url, **kwargs)

    def test_unknown_command(self):
        self.assertRaises(ValueError, HttpProxy({"timeout": 1}).run, "derp")

    def test_invalid_host(self):
        proxy = HttpProxy({"host": "http://weird.foodomain.nope", "timeout": 5})

        with self.assertRaises(CommunicationError) as cm:
            proxy.run("get", pocket=2000, eid=1)

        error_message = cm.exception.args[0]
        self.assertTrue(error_message.startswith("Error sending request: "))
        self.assertIn("NewConnectionError", error_message)


if __name__ == "__main__":
    unittest.main()



import time
import unittest

from financeager import config as cfg
from financeager import exceptions

from financeager_flask import main


class ConfigTestCase(unittest.TestCase):
    def test_load_custom_config_file(self):
        # create custom config file, modify service name
        filepath = "/tmp/{}".format(int(time.time()))
        with open(filepath, "w") as file:
            file.write("[SERVICE]\nname = flask\n")

        config = cfg.Configuration(filepath=filepath, plugins=[main.main()])
        self.assertEqual(config.get_option("SERVICE", "name"), "flask")

    def test_invalid_config(self):
        filepath = "/tmp/{}".format(int(time.time()))

        for content in ("[SERVICE:FLASK]\ntimeout = foo\n",):
            with open(filepath, "w") as file:
                file.write(content)
            self.assertRaises(
                exceptions.InvalidConfigError,
                cfg.Configuration,
                filepath=filepath,
                plugins=[main.main()],
            )


if __name__ == "__main__":
    unittest.main()



# This is an examplatory WSGI configuration file. I use it to run financeager on
# pythonanywhere.
# It is assumed that the financeager project is cloned to the home directory and
# all requirements are installed.

import os
import sys

from financeager_flask import DATA_DIR
from financeager_flask.flask import create_app

path = os.path.expanduser("~/financeager-flask/financeager_flask/flask.py")
if path not in sys.path:
    sys.path.append(path)

# the 'application' object will be used by the WSGI server
application = create_app(
    data_dir=DATA_DIR,
    config={"DEBUG": True},
)



"""Expections of the financeager-flask plugin."""
from financeager.exceptions import FinanceagerException


class OfflineRecoveryError(FinanceagerException):
    """Error during recovering of offline backup."""



import os

from financeager import DATA_DIR

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

OFFLINE_FILEPATH = os.path.join(DATA_DIR, "offline.json")

# URL endpoints
POCKETS_TAIL = "/pockets"
COPY_TAIL = "/copy"
VERSION_TAIL = "/version"

# HTTP communication defaults
DEFAULT_HOST = "http://127.0.0.1:5000"
DEFAULT_TIMEOUT = 10


def version(package_name=__package__):
    return importlib_metadata.version(package_name)



"""Utilities to create flask webservice."""
import os

from financeager import (
    init_logger,
    make_log_stream_handler_verbose,
    server,
    setup_log_file_handler,
)
from flask import Flask
from flask_restful import Api

from . import COPY_TAIL, POCKETS_TAIL, VERSION_TAIL, resources

logger = init_logger(__name__)


def create_app(data_dir=None, config=None):
    """Create web app with RESTful API built from resources. The function is
    named such that the flask cli detects it as app factory method.
    The log file handler is set up very first.
    If 'data_dir' or the environment variable 'FINANCEAGER_FLASK_DATA_DIR' is
    given, a directory is created to store application data.
    An instance of 'server.Server' is created, passing 'data_dir'. If 'data_dir'
    is not given, the application data is stored in memory and will be lost when
    the app terminates.
    'config' is a dict of configuration variables that flask understands.
    """
    setup_log_file_handler()

    # Propagate flask and werkzeug log messages to financeager logs
    init_logger("flask.app")
    init_logger("werkzeug")

    app = Flask(__name__)
    app.config.update(config or {})
    if app.debug:
        make_log_stream_handler_verbose()

    data_dir = data_dir or os.environ.get("FINANCEAGER_FLASK_DATA_DIR")
    if data_dir is None:
        logger.warning(
            "'data_dir' not given. Application data is stored in "
            "memory and is lost when the flask app terminates. Set "
            "the environment variable FINANCEAGER_FLASK_DATA_DIR "
            "accordingly for persistent data storage."
        )
    else:
        os.makedirs(data_dir, exist_ok=True)

    logger.debug(
        "Created flask app {} - {} mode".format(
            app.name, "debug" if app.debug else "production"
        )
    )

    srv = server.Server(data_dir=data_dir)
    logger.debug("Started financeager server with data dir '{}'".format(data_dir))

    api = Api(app)
    api.add_resource(
        resources.PocketsResource, POCKETS_TAIL, resource_class_args=(srv,)
    )
    api.add_resource(resources.CopyResource, COPY_TAIL, resource_class_args=(srv,))
    api.add_resource(
        resources.VersionResource, VERSION_TAIL, resource_class_args=(srv,)
    )
    api.add_resource(
        resources.PocketResource,
        "{}/<pocket_name>".format(POCKETS_TAIL),
        resource_class_args=(srv,),
    )
    api.add_resource(
        resources.EntryResource,
        "{}/<pocket_name>/<table_name>/<eid>".format(POCKETS_TAIL),
        resource_class_args=(srv,),
    )

    # Assign attribute such that e.g. test_cli can access Server methods
    app._server = srv

    return app



"""Construction and handling of HTTP requests to communicate with webservice."""
import http
import json

import requests
from financeager import DEFAULT_POCKET_NAME, DEFAULT_TABLE, exceptions

from . import COPY_TAIL, DEFAULT_HOST, POCKETS_TAIL, VERSION_TAIL

VERSION_MESSAGE = (
    "The webserver runs financeager-flask {version}\n"
    + "               and financeager       {financeager_version}"
)


class Proxy:
    """Proxy for communicating with webservice via HTTP."""

    def __init__(self, http_config=None):
        """Args:
        http_config (dict): HTTP configuration with fields 'host' (default:
            DEFAULT_HOST), 'timeout' (default: DEFAULT_TIMEOUT) and
            optionally 'username'/'password' (for basic auth)
        """
        self.http_config = http_config or {}

    def run(self, command, **data):
        """Convert specified command and data into HTTP request, send it to
        webservice, and return response. Handle error responses.
        The data kwargs are passed to the HTTP request.
        'pocket' and 'table_name' data fields are substituted, if None.

        :return: dict. See Server class for possible keys
        :raise: ValueError if invalid command given
        :raise: CommunicationError on e.g. timeouts or server-side errors,
            InvalidRequest on invalid requests
        """

        pocket = data.pop("pocket", None) or DEFAULT_POCKET_NAME

        host = self.http_config.get("host", DEFAULT_HOST)
        base_url = "{}{}".format(host, POCKETS_TAIL)
        pocket_url = "{}/{}".format(base_url, pocket)
        copy_url = "{}{}".format(host, COPY_TAIL)
        version_url = "{}{}".format(host, VERSION_TAIL)
        eid_url = "{}/{}/{}".format(
            pocket_url, data.get("table_name") or DEFAULT_TABLE, data.get("eid")
        )

        username = self.http_config.get("username")
        password = self.http_config.get("password")
        auth = None
        if username and password:
            auth = (username, password)

        kwargs = dict(auth=auth, timeout=self.http_config.get("timeout"))

        if command == "list":
            # Correctly send filters; allowing for server-side deserialization
            kwargs["json"] = json.dumps(data)
        else:
            kwargs["json"] = data or None

        if command == "list":
            url = pocket_url
            function = requests.get
        elif command == "remove":
            url = eid_url
            function = requests.delete
        elif command == "add":
            url = pocket_url
            function = requests.post
        elif command == "pockets":
            url = base_url
            function = requests.post
        elif command == "copy":
            url = copy_url
            function = requests.post
        elif command == "get":
            url = eid_url
            function = requests.get
        elif command == "update":
            url = eid_url
            function = requests.patch
        elif command == "web-version":
            url = version_url
            function = requests.get
        else:
            raise ValueError("Unknown command: {}".format(command))

        try:
            response = function(url, **kwargs)
        except requests.RequestException as e:
            raise exceptions.CommunicationError("Error sending request: {}".format(e))

        if response.ok:
            if command == "web-version":
                return VERSION_MESSAGE.format(**response.json())
            return response.json()
        else:
            try:
                # Get further information about error (see Server.run)
                error = response.json()["error"]
            except (json.JSONDecodeError, KeyError):
                error = "-"

            status_code = response.status_code
            if 400 <= status_code < 500:
                error_class = exceptions.InvalidRequest
            else:
                error_class = exceptions.CommunicationError

            message = (
                "Error handling request. "
                + "Server returned '{} ({}): {}'".format(
                    http.HTTPStatus(status_code).phrase, status_code, error
                )
            )

            raise error_class(message)



import financeager
from financeager import clients
from financeager import exceptions as base_exceptions
from financeager import plugin

from . import DEFAULT_HOST, DEFAULT_TIMEOUT, exceptions, httprequests, offline

SERVICE_NAME = "flask"
CONFIG_SECTION_NAME = "SERVICE:{}".format(SERVICE_NAME.upper())


class _Configuration(plugin.PluginConfiguration):
    def init_defaults(self, config_parser):
        config_parser[CONFIG_SECTION_NAME] = {
            "host": DEFAULT_HOST,
            "timeout": DEFAULT_TIMEOUT,
            "username": "",
            "password": "",
        }

    def init_option_types(self, option_types):
        option_types[CONFIG_SECTION_NAME] = {
            "timeout": "int",
        }


class _CliOptions(plugin.PluginCliOptions):
    def extend(self, command_parser):
        command_parser.add_parser(
            "web-version",
            help="information about financeager versions installed on server "
            "(only if configured with service name '{}')".format(SERVICE_NAME),
        )


class _Client(clients.Client):
    """Client for communicating with the financeager Flask webservice."""

    def __init__(self, *, configuration, sinks):
        """Set up proxy and urllib3 logger."""
        super().__init__(configuration=configuration, sinks=sinks)
        self.proxy = httprequests.Proxy(
            http_config=configuration.get_section(CONFIG_SECTION_NAME)
        )

        financeager.init_logger("urllib3")

    def safely_run(self, command, **params):
        """Execute base functionality.
        If successful, attempt to recover offline backup. Otherwise store
        request in offline backup.
        Return whether execution was successful.

        :return: bool
        """
        success = super().safely_run(command, **params)

        if success:
            try:
                # Avoid recursion by passing base class for invoking safely_run
                if offline.recover(super()):
                    self.sinks.info("Recovered offline backup.")

            except exceptions.OfflineRecoveryError:
                self.sinks.error("Offline backup recovery failed!")
                success = False

        # If request was erroneous, it's not supposed to be stored offline
        if (
            not isinstance(self.latest_exception, base_exceptions.InvalidRequest)
            and self.latest_exception is not None
            and offline.add(command, **params)
        ):
            self.sinks.info("Stored '{}' request in offline backup.".format(command))

        return success


def main():
    return plugin.ServicePlugin(
        name=SERVICE_NAME,
        client=_Client,
        config=_Configuration(),
        cli_options=_CliOptions(),
    )
